#
# This script is part of TkKasse. You may use, modify and distribute it under the terms
# of the GNU General Public License, Version 2.
# See details at http://www.fsf.org/licenses/gpl.txt
# (C) by Jan Kandziora <tkkasse@users.sf.net> 
#


## Generic connection class:
itcl::class Connection \
{
	constructor { SERVER_ARG FD_ARG REMOTE_HOST_ARG REMOTE_PORT_ARG } {}
	destructor {}

	method eval { args }

	protected \
	{
		variable server {}
		variable fd {}
	}
}

## Generic connection constructor:
itcl::body Connection::constructor { SERVER_ARG FD_ARG REMOTE_HOST_ARG REMOTE_PORT_ARG } \
{
	## Save the server argument.
	set server $SERVER_ARG

	## Save the file descriptor argument.
	set fd $FD_ARG
}

## Generic connection destructor:
itcl::body Connection::destructor {} \
{
	## Close the descriptor.
	close $fd
}

## Eval args in the scope of this connection:
itcl::body Connection::eval { args } \
{
	::eval ::eval $this $args
}



## Line connection class:
itcl::class Line::Connection \
{
	inherit ::Connection

	constructor { SERVER_ARG FD_ARG REMOTE_HOST_ARG REMOTE_PORT_ARG } { chain $SERVER_ARG $FD_ARG $REMOTE_HOST_ARG $REMOTE_PORT_ARG } {}

	method processLine {}
}

## Line connection constructor:
itcl::body Line::Connection::constructor { SERVER_ARG FD_ARG REMOTE_HOST_ARG REMOTE_PORT_ARG } \
{
	## Configure the socket for line-by-line operation.
	fconfigure $fd \
		-buffering line \
		-blocking false 

	## Register a handler for read events.
	fileevent $fd readable "if { \[ itcl::find objects $this \] != {} } { $this processLine }"
}

## Process a line of text:
## NOTE: This is a callback and not intended as a user callable method.
itcl::body Line::Connection::processLine {} \
{
	## Read the line.
	if { [ gets $fd LINE ] <= 0 } \
	{
		## There was a problem. Check if the line could be read completely.
		if [ fblocked $fd ] \
		{
			## No. In that case LINE was set to {} by gets, and no input was consumed.
			## Do nothing special. The next fileevent will get us more data.
		} \
		{
			## Yes. The peer has closed the connection, or there was a network failure
			## In any case, the server has to silently close its part of the connection now.
			$server closeConnection $this
		}
	}

	## Return the read data.
	return $LINE
}



## Command connection class:
itcl::class Command::Connection \
{
	inherit Line::Connection

	constructor { SERVER_ARG FD_ARG REMOTE_HOST_ARG REMOTE_PORT_ARG } { chain $SERVER_ARG $FD_ARG $REMOTE_HOST_ARG $REMOTE_PORT_ARG } {}
	destructor {}

	method setErrorHandler { ERROR_HANDLER_ARG }
	method processLine {}

	protected \
	{
		variable error_handler {}
		variable interpreter {}
	}
}

## Command connection constructor:
itcl::body Command::Connection::constructor { SERVER_ARG FD_ARG REMOTE_HOST_ARG REMOTE_PORT_ARG } \
{
	## Create a safe interpreter for this connection.
	set interpreter [ interp create -safe ]

	## Create a command alias for remote error handling.
	$interpreter alias setErrorHandler $this setErrorHandler
}

## Command connection destructor:
itcl::body Command::Connection::destructor {} \
{
	## Remove the safe interpreter.
	interp delete $interpreter

	## Call the base class destructor.
	chain
}

## Set/Unset the remote error handler:
itcl::body Command::Connection::setErrorHandler { ERROR_HANDLER_ARG } \
{
	## The following substitutions are possible:
	## %c Tcl error code
	## %i Error info
	## %d Error time
	## %m error message
	## %C Errorneous command

	set error_handler $ERROR_HANDLER_ARG

	log debug "Error handler set for connection $this : $error_handler"
}

## Process a line of text:
itcl::body Command::Connection::processLine {} \
{
	## Read the line.
	set LINE [ chain ]
	
	## Only interpret if a line was given.
	## If there was no input, due to an incomplete line in the input buffer or a closed connection, do nothing.
	if { $LINE != "" } \
	{
		## Put the line into the interpreter.
		if [ catch { $interpreter eval $LINE } ERROR_MESSAGE ] \
		{
			global errorCode
			global errorInfo

			## Look if an error handler was given.
			if { $error_handler != "" } \
			{
				## Do the substitutions in the error handler.
				set HANDLER $error_handler

				## %c in the handler call is replaced by the tcl error code.
				regsub -all {\\|&} [ list $errorCode ] {\\\0} SUBST
				regsub -all "%c" $HANDLER $SUBST HANDLER

				## %i in the handler call is replaced by the tcl error info.
				regsub -all {\\|&} [ list $errorInfo ] {\\\0} SUBST
				regsub -all \n $SUBST "" SUBST
				regsub -all "%i" $HANDLER $SUBST HANDLER

				## %D in the handler call is replaced by the time of the error.
				regsub -all {\\|&} [ clock seconds ] {\\\0} SUBST
				regsub -all "%d" $HANDLER $SUBST HANDLER

				## %m in the handler call is replaced by the error message.
				regsub -all {\\|&} [ list $ERROR_MESSAGE ] {\\\0} SUBST
				regsub -all \n $SUBST "" SUBST
				regsub -all "%m" $HANDLER $SUBST HANDLER

				## %C in the handler call is replaced by the errorneous command.
				regsub -all {\\|&} [ list $LINE ] {\\\0} SUBST
				regsub -all \n $SUBST "" SUBST
				regsub -all "%C" $HANDLER $SUBST HANDLER

				## Send the handler command to the client.
				puts $fd $HANDLER
			}

			## Log errors.
			if { [ lindex $errorCode 0 ] == "TKKASSE" } \
			{
				## Log tkkasse internal errors only to debug log.
				log debug $errorInfo
			} \
			{
				## Log all other errors as critical.
				log crit "CRITICAL: $errorInfo"
			}

		}
	}
}


	
## Authenticated connection class:
itcl::class Authenticated::Connection \
{
	inherit Command::Connection

	constructor { SERVER_ARG FD_ARG REMOTE_HOST_ARG REMOTE_PORT_ARG } { chain $SERVER_ARG $FD_ARG $REMOTE_HOST_ARG $REMOTE_PORT_ARG } {}

	method authenticateUser { USER_LOGIN_NAME_ARG USER_PASSWORD_ARG }
	method goAnonymous {}
	method processLine {}
	method testUserRights { args }

	protected \
	{
		variable user {}
		variable user_rights_cache {}
	}
	
}

## Authenticated connection constructor:
itcl::body Authenticated::Connection::constructor { SERVER_ARG FD_ARG REMOTE_HOST_ARG REMOTE_PORT_ARG } \
{
	## Create command aliases for authentication and anonymous mode.
	$interpreter alias authenticateUser $this authenticateUser
	$interpreter alias goAnonymous $this goAnonymous
}

## Authenticate an user:
itcl::body Authenticated::Connection::authenticateUser { USER_LOGIN_NAME_ARG USER_PASSWORD_ARG } \
{
	## Get the object descriptor for this user.
	## The server calls its user manager to check this. It throws an error
	## if the user cannot be authenticated due to nonexistence or wrong password.
	set user [ $server authenticateUser $USER_LOGIN_NAME_ARG $USER_PASSWORD_ARG ]

	## The user is now authenticated.
	## The user rights are updated before processing the next command.
}

## Go to anonymous mode:
itcl::body Authenticated::Connection::goAnonymous {} \
{
	## Clear the user object descriptor variable.
	set user {}

	## The user is now anonymous.
	## The user rights are updated before processing the next command.
}

## Process a line of text:
itcl::body Authenticated::Connection::processLine {} \
{
	## Check if this is an anonymous connection.
	if { $user != {} } \
	{
		## No. Update the user's rights cache. The cache ensures consistent rights until the line is processed completely.
		## Check if the user object was deleted.
		if [ catch { $user cget -rights } user_rights_cache ] \
		{
			## Yes. Go to anonymous mode.
			goAnonymous

			## Do not process the line. Complain to the client due to an deleted user.
			log warn "User on connection $this was deleted. Logoff forced."
			return -code error -errorcode { TKKASSE USER UNKNOWN } "user was deleted. logoff forced."
		}
	} \
	{
		## Yes. Anonymous users have no special rights.
		set user_rights_cache {}
	}

	## Process the line.
	chain
}

## Test a user's rights to do something.
itcl::body Authenticated::Connection::testUserRights { args } \
{
	## All the noted rights must be given in the rights cache.
	foreach ARG $args \
	{
		if { [ lsearch -exact $user_rights_cache $ARG ] == -1 } \
		{
			## The right is not given. Fail.
			return -code error -errorcode { TKKASSE USER RIGHTS_INSUFFICIENT } "insufficient rights: $ARG is neccessary to do the desired action."
		}
	}

	## All noted rights are given. Success.
}

