#
# This script is part of TkKasse. You may use, modify and distribute it under the terms
# of the GNU General Public License, Version 2.
# See details at http://www.fsf.org/licenses/gpl.txt
# (C) by Jan Kandziora <tkkasse@users.sf.net> 
#


## Evaluate in the TkKasse namespace.
namespace eval TkKasse \
{
	## TextOnly printer driver class. Prints just text with configureable width.
	itcl::class PrinterDriver::TextOnly \
	{
		inherit ::TkKasse::PrinterDriver
	
		constructor { args } { eval chain $args } {}
		
		method convert { PRINT_DESCRIPTION_ARG }
		method convertPrimitives { PRINT_PRIMITIVES_ARG }
		method print { ITEM_OBJECT_ARG }

		public \
		{
			variable autowrap 0
			variable encoding [ encoding system ]
			variable form_feed "\n************************cut***************************\n"
			variable init "" 
			variable newline \x0a
			variable space " "
			variable paper_width 40
			variable paper_height * 
		}

		protected \
		{
			variable interpreter
			variable raw_print_data
		}
	}

	## TextOnly printer driver class constructor.
	itcl::body PrinterDriver::TextOnly::constructor { args } \
	{
		## Create interpreter to understand the print description.
		## TODO: This interpreter is unsafe by now to make package loading work!
		## That's not important as long as the printer driver data is given from within
		## this server program, but be careful not to use user data as printer driver input
		## ::safe::interCreate won't work also, have to find out why!
		set interpreter [ interp create ]

		## Create commands within the slave interpreter.
		$interpreter alias formatCurrency formatCurrency
		$interpreter alias getCurrencySymbol getCurrencySymbol
		$interpreter alias log log
		$interpreter alias print $this print
		$interpreter eval \
		{
			package require Itcl
			package require textutil

			## Generic item class:
			itcl::class item \
			{
				constructor { args } { eval configure $args } {}

				## Abstract item drawing method. It has to return a list of primitives to draw.
				## For this TextOnly driver, the primitives are character strings and their
				## position is accurate within one single char width and height.
				method draw {}

				public \
				{
					## Height and width of the item.
					variable height 0
					variable width 0
				}
			}



			## Fill item class:
			itcl::class item::fill \
			{
				inherit item

				constructor { args } { eval chain $args } {}
				
				method draw {}

				public \
				{
					## Fill char.
					variable char {}

					## Height and width of the item.
					variable height 0
					variable width 0
				}
			}

			## Change the fill according to the given size.
			itcl::configbody item::fill::height \
			{
				## Check the height for the fill.
				if { $height == "*" } \
				{
					## It is given as *, the fill alone has a height of zero.
					set height 0
				}
			}

			## Draw the fill item:
			itcl::body item::fill::draw {} \
			{
				## Fill the whole area line by line.
				set PRIMITIVES {}
				for { set Y 0 } { $Y < $height } { incr Y } \
				{
					lappend PRIMITIVES [ list 0 $Y [ string repeat $char $width ] ]
				}

				return $PRIMITIVES
			}



			## Pad item class:
			itcl::class item::pad \
			{
				inherit item

				constructor { args } { eval chain $args } {}
				
				method draw {}

				public \
				{
					## Slave item within the padding.
					variable slave {}

					## Padding. 
					variable padx 0
					variable pady 0

					## Height and width of the item.
					variable height 0
					variable width 0
				}
			}

			## Change the size of the slave according the own size.
			itcl::configbody item::pad::width \
			{
				$slave configure -width [ expr $width - $padx * 2 ]
			}
			
			itcl::configbody item::pad::height \
			{
				## Check the height for the pad.
				if { $height == "*" } \
				{
					## It is given as *, the pad has the height of the slave + pady*2.
					set height [ expr [ $slave cget -height ] + $pady * 2 ]
				} \
				{
					$slave configure -height [ expr $height - $pady * 2 ]
				}
			}

			## Calculate the size of the pad item.
			itcl::configbody item::pad::padx \
			{
				set width [ expr [ $slave cget -width ] + $padx * 2 ]
			}

			itcl::configbody item::pad::pady \
			{
				set height [ expr [ $slave cget -height ] + $pady * 2 ]
			}
			
			itcl::configbody item::pad::slave \
			{
				set width [ expr [ $slave cget -width ] + $padx * 2 ]
				set height [ expr [ $slave cget -height ] + $pady * 2 ]
			}

			## Draw the pad item:
			itcl::body item::pad::draw {} \
			{
				set PRIMITIVES {}
				foreach PRIMITIVE [ $slave draw ] \
				{
					## Draw the primitive of the slave item at the position specified by padx and pady.
					lappend PRIMITIVES [ list [ expr [ lindex $PRIMITIVE 0 ] + $padx ] [ expr [ lindex $PRIMITIVE 1 ] + $pady ] [ lindex $PRIMITIVE 2 ] ]
				}

				## Return the moved primitives. 
				return $PRIMITIVES
			}



			## String item class:
			itcl::class item::string \
			{
				inherit item

				constructor { args } { eval chain $args } {}
				
				method draw {}

				public \
				{
					## Position: left, right, center
					variable position {} 

					## String to draw.
					variable data {} 

					## Height and width of the item.
					variable height 0
					variable width 0
				}
			}

			## Recalculate the width anytime the data changes.
			itcl::configbody item::string::data \
			{
				set width [ string length $data ]
				if { $data == {} } { set height 0 } { set height 1 }
			}

			## Change the string according to the given size.
			itcl::configbody item::string::height \
			{
				## Check the height for the string.
				if { $height == "*" } \
				{
					## It is given as *, the string alone has a height of one.
					set height 1
				}
			}

			## Draw the string item.
			itcl::body item::string::draw {} \
			{
				set LINE_LENGTH [ string length $data ]

				## Check for left, right and center position.
				switch -exact $position \
				{
					center - c \
					{
						## Horizontal centered.
						if { $width > $LINE_LENGTH } \
						{
							## Center the string.
							lappend PRIMITIVES [ list [ expr round( ( ${width}.0 - ${LINE_LENGTH}.0 ) / 2) ] 0 $data ]
						} \
						{
							## The string width exceeds the bounding box width, cut the string equally at both sides.
							set LINE_START [ expr round( ( ${LINE_LENGTH}.0 - ${width}.0 ) / 2) ]
							set LINE_END [ expr $LINE_START + $width - 1 ]
							lappend PRIMITIVES [ list 0 0 [ string range $data $LINE_START $LINE_END ] ]
						}
					}
					{} - left - l \
					{
						## Left side.
						if { $width > $LINE_LENGTH } \
						{
							## Put the string to the left side.
							lappend PRIMITIVES [ list 0 0 $data ]
						} \
						{
							## Cut the string on the right side.
							lappend PRIMITIVES [ list 0 0 [ string range $data 0 [ expr $width - 1 ] ] ]
						}
					} 
					right - r \
					{
						## Right side.
						if { $width > $LINE_LENGTH } \
						{
							## Put the string to the right side.
							lappend PRIMITIVES [ list [ expr $width - $LINE_LENGTH ] 0 $data ]
						} \
						{
							## Cut the string on the left side.
							lappend PRIMITIVES [ list 0 0 [ string range $data [ expr $LINE_LENGTH - $width ] end ] ]
						}
					} 
				}

	 			## Return the primitives list.
				return $PRIMITIVES
			}



			## Table item class: 
			itcl::class item::table \
			{
				inherit item

				constructor { args } { eval chain $args } {}
			
				method draw {}

				public \
				{
					## List of slave items and their properties within the table.
					variable slaves {}

					## List of column properties. The number of elements determines the number of columns.
					variable columns {}

					## List of row properties.
					variable rows {}

					## Height and width of the item.
					variable height 0
					variable width 0
				}
			}

			## Change the size of the slaves according to the own size.
			itcl::configbody item::table::width \
			{
				## Get some values before calculating the actual width of each column.
				set FILL $width
				array set COLUMN_WIDTH {}
				set ADDED_WEIGHT 0
				set INDEX 0
				set COLUMN_WEIGHTS {}
				foreach COLUMN $columns \
				{
					## Get the fill.
					set MINIMUM_WIDTH [ lindex $COLUMN 0 ]
					set FILL [ expr $FILL - $MINIMUM_WIDTH ]

					## Get the initial column width.
					set COLUMN_WIDTH($INDEX) $MINIMUM_WIDTH

					## Get the added weight value.
					set WEIGHT [ lindex $COLUMN 1 ]
					set ADDED_WEIGHT [ expr $ADDED_WEIGHT + $WEIGHT ]

					## Set up the local column list.
					if { $WEIGHT > 0 } \
					{
						lappend COLUMN_WEIGHTS [ list $INDEX $WEIGHT ]
					}

					## Next column.
					incr INDEX
				}

				## Calculate the actual column width from the column weight.
				set FILL_OLD $FILL 
				foreach COLUMN $COLUMN_WEIGHTS \
				{
					## Make the first guess with float numbers rounded down.
					set COLUMN_FILL [ expr ($FILL_OLD * [ lindex $COLUMN 1 ]) / $ADDED_WEIGHT ]

					## Recalculate the column width and the fill still left.
					set COLUMN_WIDTH([ lindex $COLUMN 0 ]) [ expr $COLUMN_WIDTH([ lindex $COLUMN 0 ]) + $COLUMN_FILL ]
					set FILL [ expr $FILL - $COLUMN_FILL ]
				}

				## The following actions depend on the sign of fill.
				if { $FILL >= 0 } \
				{
					## The fill is positive. We have to expand the expandable columns.

					## The maximum fill possibly left is the same as number of columns with a weight > 0.
					## Give the fill to the columns sorted by weight.
					foreach COLUMN [ lsort -index 1 -integer -decreasing $COLUMN_WEIGHTS ] \
					{
						## Break if no fill is left.
						if { $FILL == 0 } break;

						## Otherwise increment the column width. Decrement the fill.
						incr COLUMN_WIDTH([ lindex $COLUMN 0 ])
						set FILL [ expr $FILL - 1 ]
					}
				} \
				{
					## The fill is negative. We have to shorten the expandable columns.

					## The maximum fill possible to steal is the same as number of columns with a weight > 0.
					## Give the fill to the columns sorted by weight.
					foreach COLUMN [ lsort -index 1 -integer -decreasing $COLUMN_WEIGHTS ] \
					{
						## Break if no fill is left.
						if { $FILL == 0 } break;

						## Otherwise increment the column width. Decrement the fill.
						set COLUMN_WIDTH([ lindex $COLUMN 0 ]) [ expr $COLUMN_WIDTH([ lindex $COLUMN 0 ]) - 1 ]
						incr FILL
					}
				}
			
				## Now we have the exact width of each column and no fill left.
				
				## Determine the number of columns.
				set COLUMN_COUNT [ llength $columns ]

				## Go through all slaves and change their width.
				set COLUMN_INDEX 0
				foreach SLAVE $slaves \
				{
					## Resize the slave width.
					$SLAVE configure -width $COLUMN_WIDTH($COLUMN_INDEX)
					
					## Next slave.
					incr COLUMN_INDEX
					if { $COLUMN_INDEX == $COLUMN_COUNT } \
					{
						## Next row.
						set COLUMN_INDEX 0
					}
				}

			}
			
			itcl::configbody item::table::height \
			{
				## Check the height for the table.
				if { $height == "*" } \
				{
					## Calculate the height, it is given as *.
					set height 0
					foreach ROW $rows \
					{
						set height [ expr $height + [ lindex $ROW 0 ] ]
					}
				} \
				{
					## TODO: Change the table layout if the height is given as a number.
				}

				## Now we have the exact height of each row and no fill left.

				## Determine the number of columns.
				set COLUMN_COUNT [ llength $columns ]

				## Go through all slaves and change their width.
				set COLUMN_INDEX 0
				set ROW_INDEX 0
				foreach SLAVE $slaves \
				{
					## Resize the slave height.
					$SLAVE configure -height [ lindex [ lindex $rows $ROW_INDEX ] 0 ]
					
					## Next slave.
					incr COLUMN_INDEX
					if { $COLUMN_INDEX == $COLUMN_COUNT } \
					{
						## Next row.
						set COLUMN_INDEX 0
						incr ROW_INDEX
					}
				}
			}

			## Recalculate the size of the columns and rows according to the slaves within.
			itcl::configbody item::table::slaves \
			{
				## Determine the number of columns.
				set COLUMN_COUNT [ llength $columns ]

				## Go through all slaves and determine the minimum width and height of each cell.
				set COLUMN_INDEX 0
				set ROW_INDEX 0
				array set MINIMUM_WIDTH {}
				array set MINIMUM_HEIGHT {0 0}

				## Process all slaves, if any.
				foreach SLAVE $slaves \
				{
					## Set minimum width and height to zero, if not set yet.
					if [ catch { set MINIMUM_WIDTH($COLUMN_INDEX) } ] { set MINIMUM_WIDTH($COLUMN_INDEX) 0 }
					if [ catch { set MINIMUM_HEIGHT($ROW_INDEX) } ] { set MINIMUM_HEIGHT($ROW_INDEX) 0 }

					## Check whether the slave is wider than the minimum width.
					if { [ $SLAVE cget -width ] > $MINIMUM_WIDTH($COLUMN_INDEX) } { set MINIMUM_WIDTH($COLUMN_INDEX) [ $SLAVE cget -width ] }
					if { [ $SLAVE cget -height ] > $MINIMUM_HEIGHT($ROW_INDEX) } { set MINIMUM_HEIGHT($ROW_INDEX) [ $SLAVE cget -height ] }

					## Next slave.
					incr COLUMN_INDEX
					if { $COLUMN_INDEX == $COLUMN_COUNT } \
					{
						## Next row.
						set COLUMN_INDEX 0
						incr ROW_INDEX
					}
				}

				## Change the minimum width of all columns with zero as minimum noted width.
				set COLUMNS {}
				foreach COLUMN_INDEX [ lsort -integer [ array names MINIMUM_WIDTH ] ] \
				{
					## Check if the row is noted.
					set COLUMN [ lindex $columns $COLUMN_INDEX ]
					if { $COLUMN == {} } \
					{
						## No. Do it now.
						lappend COLUMNS [ list $MINIMUM_WIDTH($COLUMN_INDEX) 0 ]
					} \
					{
						## Yes, check if the column has a minimum noted width of zero.
						if { [ lindex $COLUMN 0 ] == 0 } \
						{
							## Yes. Change the minimum width.
							lappend COLUMNS [ list $MINIMUM_WIDTH($COLUMN_INDEX) [ lindex $COLUMN 1 ] ]
						} \
						{
							## No. Don't change the column.
							lappend COLUMNS $COLUMN
						}
					}

					## Next column.
					incr COLUMN_INDEX
				}
				set columns $COLUMNS

				## Calculate the width. 
				set width 0
				foreach COLUMN $columns \
				{
					set width [ expr $width + [ lindex $COLUMN 0 ] ]
				}

				## Change the minimum height of all rows with zero as minimum noted height.
				set ROWS {}
				foreach ROW_INDEX [ lsort -integer [ array names MINIMUM_HEIGHT ] ] \
				{
					## Check if the row is noted.
					set ROW [ lindex $rows $ROW_INDEX ]
					if { $ROW == {} } \
					{
						## No. Do it now.
						lappend ROWS [ list $MINIMUM_HEIGHT($ROW_INDEX) 0 ]
					} \
					{
						## Yes, check if the row has a minimum noted height of zero.
						if { [ lindex $ROW 0 ] == 0 } \
						{
							## Yes. Change the minimum height.
							lappend ROWS [ list $MINIMUM_HEIGHT($ROW_INDEX) [ lindex $ROW 1 ] ]
						} \
						{
							## No. Don't change the row.
							lappend ROWS $ROW
						}
					}

					## Next row.
					incr ROW_INDEX
				}
				set rows $ROWS

				## Calculate the height. 
				set height 0
				foreach ROW $rows \
				{
					set height [ expr $height + [ lindex $ROW 0 ] ]
				}
			}
			
			## Draw the table item.
			itcl::body item::table::draw {} \
			{
				set PRIMITIVES {}

				## Determine the number of columns.
				set COLUMN_COUNT [ llength $columns ]

				## Go through all slaves.
				set X 0
				set Y 0
				set COLUMN_INDEX 0
				foreach SLAVE $slaves \
				{
					## Move the primitives of the slave to the position.
					foreach PRIMITIVE [ $SLAVE draw ] \
					{
						## Draw the primitive of the slave item at the position specified by X and Y.
						lappend PRIMITIVES [ list [ expr [ lindex $PRIMITIVE 0 ] + $X ] [ expr [ lindex $PRIMITIVE 1 ] + $Y ] [ lindex $PRIMITIVE 2 ] ]
					}

					## Next slave.
					incr COLUMN_INDEX
					set X [ expr $X + [ $SLAVE cget -width ] ]
					if { $COLUMN_INDEX == $COLUMN_COUNT } \
					{
						## Next row.
						set COLUMN_INDEX 0
						set X 0
						set Y [ expr $Y + [ $SLAVE cget -height ] ]
					}
				}

				## Return the moved primitives. 
				return $PRIMITIVES
			}



			## Text item class: 
			itcl::class item::text \
			{
				inherit item

				constructor { args } { eval chain $args } {}
				
				method draw {}

				public \
				{
					## Sticky: combination of n,s,e,w or c(center)
					variable sticky {} 

					## Text to draw.
					variable data {}

					## Line separator within the data. 
					variable line_seperator ""

					## Height and width of the item.
					variable height 0 
					variable width 0
				}
			}

			## Recalculate the width and height anytime the data changes.
			itcl::configbody item::text::data \
			{
				## Find out the longest line and the number of lines.
				set width 0
				set height 0
				foreach LINE [ split $data $line_seperator ] \
				{
					if { [ string length $LINE ] > $width } { set width [ string length $LINE ] }
					incr height
				}
			}

			## Change the text according to the given size.
			itcl::configbody item::text::height \
			{
				## Check the height for the text.
				if { $height == "*" } \
				{
					## It is given as *, the text alone has a height according to the number of lines.
					set height [ llength [ split $data $line_seperator ] ]
				}
			}

			## Draw the text item.
			itcl::body item::text::draw {} \
			{
				## Format each line for itself.
				set LINES {}
				foreach LINE [ split $data $line_seperator ] \
				{
					set LINE_LENGTH [ string length $LINE ]

					## Check for west, east and center position.
					switch -exact $sticky \
					{
						{} - c - n - s - ns - sn \
						{
							## Horizontal centered.
							if { $width > $LINE_LENGTH } \
							{
								## Center the string.
								lappend LINES [ list [ expr round( ( ${width}.0 - ${LINE_LENGTH}.0 ) / 2) ] $LINE ]
							} \
							{
								## The line width exceeds the bounding box width, cut the line equally at both sides.
								set LINE_START [ expr round( ( ${LINE_LENGTH}.0 - ${width}.0 ) / 2) ]
								set LINE_END [ expr $LINE_START + $width - 1 ]
								lappend LINES [ list 0 [ string range $LINE $LINE_START $LINE_END ] ]
							}
						}
						w - nw - sw - wn - ws - nsw - nws - snw - swn - wns - wsn \
						{
							## Left side.
							if { $width > $LINE_LENGTH } \
							{
								## Put the line to the left side.
								lappend LINES [ list 0 $LINE ]
							} \
							{
								## Cut the line on the right side.
								lappend LINES [ list 0 [ string range $LINE 0 [ expr $width - 1 ] ] ]
							}
						} 
						e - ne - se - en - es - nse - nes - sne - sen - ens - esn \
						{
							## Right side.
							if { $width > $LINE_LENGTH } \
							{
								## Put the line to the right side.
								lappend LINES [ list [ expr $width - $LINE_LENGTH ] $LINE ]
							} \
							{
								## Cut the line on the left side.
								lappend LINES [ list 0 [ string range $LINE [ expr $LINE_LENGTH - $width ] end ] ]
							}
						} 
						ew - we - nsew - nswe - nesw - news - nwse - nwes - snew - snwe - senw - sewn - swne - swen - \
						ewns - ewsn - eswn - esnw - ensw - enws - wsen - wsne - wesn - wens - wnes - wnse \
						{
							## Justify horizontally.
							if { $width > $LINE_LENGTH } \
							{
								lappend LINES [ list 0 [ textutil::adjust $LINE -justify plain -length $width ] ]
							} \
							{
								## The line width exceeds the given width, cut the line equally at both sides.
								## This is the same behaviour as center.
								set LINE_START [ expr round( ( ${LINE_LENGTH}.0 - ${WIDTH}.0 ) / 2) ]
								set LINE_END [ expr $LINE_START + $WIDTH - 1 ]
								lappend LINES [ list 0 [ string range $LINE $LINE_START $LINE_END ] ]
							}
						}
					}
				}

				## Now we have to decide about the vertical stickyness.
				set LINE_COUNT [ llength $LINES ]
				set PRIMITIVES {}
				switch -exact $sticky \
				{
					{} - c - e - w - ew - we \
					{
						## Vertically centered.
						if { $height > $LINE_COUNT } \
						{
							## Center vertically.
							set Y [ expr round( ( ${height}.0 - ${LINE_COUNT}.0 ) / 2) ]

							## Compose the primitive list.
							foreach LINE $LINES \
							{
								lappend PRIMITIVES [ list [ lindex $LINE 0 ] $Y [ lindex $LINE 1 ] ]
								incr Y
							}
						} \
						{
							## The line count exceeds the given height, throw away lines equally at both sides.
							set START_LINE [ expr round( ( ${LINE_COUNT}.0 - ${height}.0 ) / 2) ]
							set END_LINE [ expr $START_LINE + $height - 1 ]
							set LINES_OUTPUT [ lrange $LINES $START_LINE $END_LINE ]

							## Start at first line of the bounding box.
							set Y 0

							## Compose the primitive list.
							foreach LINE $LINES_OUTPUT \
							{
								lappend PRIMITIVES [ list [ lindex $LINE 0 ] $Y [ lindex $LINE 1 ] ]
								incr Y
							}
						}
					}
					n - ne - en - nw - wn - new - nwe - ewn - enw - wen - wne \
					{
						## Top.
						if { $height > $LINE_COUNT } \
						{
							## Start at first line of the bounding box.
							set Y 0

							## Compose the primitive list.
							foreach LINE $LINES \
							{
								lappend PRIMITIVES [ list [ lindex $LINE 0 ] $Y [ lindex $LINE 1 ] ]
								incr Y
							}
						} \
						{
							## Throw away some lines at the bottom.
							set LINES_OUTPUT [ lrange $LINES 0 [ expr $height - 1  ] ]

							## Start at first line of the bounding box.
							set Y 0

							## Compose the primitive list.
							foreach LINE $LINES_OUTPUT \
							{
								lappend PRIMITIVES [ list [ lindex $LINE 0 ] $Y [ lindex $LINE 1 ] ]
								incr Y
							}
						}
					}
					s - se - es - sw - ws - sew - swe - ews - esw - wes - wse \
					{
						## Bottom.
						if { $height > $LINE_COUNT } \
						{
							## Put the line at the bottom of the bounding box.
							set Y [ expr $height - $LINE_COUNT ]

							## Compose the primitive list.
							foreach LINE $LINES \
							{
								lappend PRIMITIVES [ list [ lindex $LINE 0 ] $Y [ lindex $LINE 1 ] ]
								incr Y
							}
						} \
						{
							## Throw away some lines at the top.
							set LINES_OUTPUT [ lrange $LINES [ expr $LINE_COUNT - $height  ] end ]

							## Start at first line of the bounding box.
							set Y 0

							## Compose the primitive list.
							foreach LINE $LINES_OUTPUT \
							{
								lappend PRIMITIVES [ list [ lindex $LINE 0 ] $Y [ lindex $LINE 1 ] ]
								incr Y
							}
						}
					}
					ns - sn - nsew - nswe - nesw - news - nwse - nwes - snew - snwe - senw - sewn - swne - swen - \
					ewns - ewsn - eswn - esnw - ensw - enws - wsen - wsne - wesn - wens - wnes - wnse \
					{
						return -code error -errorcode { TKKASSE TODO } "function not implemented yet."
						## TODO: Implement!
						## Justify lines.
					}
				}

	 			## Return the primitives list.
				return $PRIMITIVES
			}
		}
	} 

	## Convert the print description into the printer language.
	itcl::body PrinterDriver::TextOnly::convert { PRINT_DESCRIPTION_ARG } \
	{
		## Set up the new raw print data.
		set raw_print_data {}

		## Delete all variables inside the interpreter.
		## Create some needed variables after that.
		$interpreter eval \
		{
			eval unset [ info vars ]
			variable errorCode errorInfo
		}

		## Preset some variabled inside the print description interpreter.
		set INDEX 0
 		set VARIABLES [ lindex $PRINT_DESCRIPTION_ARG 0 ]
 		set VALUES [ lindex $PRINT_DESCRIPTION_ARG 1 ]
		foreach VARIABLE $VARIABLES \
		{
			$interpreter eval "set $VARIABLE \"[ lindex $VALUES $INDEX ]\""
			incr INDEX
		}

		## Interprete the print description.
		$interpreter eval "[ lindex $PRINT_DESCRIPTION_ARG 2 ]"

		## Delete all items created inside the interpreter during the print description run.
		$interpreter eval \
		{
			eval itcl::delete object [ itcl::find objects -isa item ]
		}


		## Return the raw printer data.
		return $raw_print_data
	} 

	## Interprete the set of primitives given by the driver and alter the raw print data.
	itcl::body PrinterDriver::TextOnly::convertPrimitives { PRINT_PRIMITIVES_ARG } \
	{
		## Set the raw print data to the printer init string.
		set RAW_PRINT_DATA $init

		## Sort print primitives by increasing line number. 
		set PRINT_PRIMITIVES [ lsort -index 1 -integer $PRINT_PRIMITIVES_ARG ]
		log debug "Print: $PRINT_PRIMITIVES"

		## Set some start values.
		set LINE [ string repeat $space $paper_width ]
		set LINE_NUMBER [ lindex [ lindex $PRINT_PRIMITIVES 0 ] 1 ]

		## Go to the first line.
		append RAW_PRINT_DATA [ string repeat $newline $LINE_NUMBER ]

		## Go through all primitives.
		foreach PRINT_PRIMITIVE $PRINT_PRIMITIVES \
		{
			## Check for a new line number.
			set OLD_LINE_NUMBER $LINE_NUMBER
			set LINE_NUMBER [ lindex $PRINT_PRIMITIVE 1 ]
			if { $LINE_NUMBER != $OLD_LINE_NUMBER } \
			{
				## New line. Output the current line, then a suitable number of newline characters.
				append RAW_PRINT_DATA [ encoding convertto $encoding $LINE ]

				## Decide whether the printer automatically wraps if the line width is used completely.
				if $autowrap \
				{
					## Automatical wrap. Leave out one newline char.
					append RAW_PRINT_DATA [ string repeat $newline [ expr $LINE_NUMBER - $OLD_LINE_NUMBER - 1 ] ]
				} \
				{
					## No automatical wrap.
					append RAW_PRINT_DATA [ string repeat $newline [ expr $LINE_NUMBER - $OLD_LINE_NUMBER ] ]
				}

				## Clear the current line.
				set LINE [ string repeat $space $paper_width ]
			}

			## Put the primitive into this line.
			set POS [ lindex $PRINT_PRIMITIVE 0 ]
			set STRING [ lindex $PRINT_PRIMITIVE 2 ]
			set LINE [ string replace $LINE $POS [ expr $POS + [ string length $STRING ] - 1 ] $STRING ]
		}
	
		## Output the last line, if there was any primitive to print.
		if { $PRINT_PRIMITIVES != {} } \
		{
			append RAW_PRINT_DATA [ encoding convertto $encoding $LINE ]
		}

		## Form feed/cut the paper.
		append RAW_PRINT_DATA $form_feed

		## Return the data.
		return $RAW_PRINT_DATA
	}

	## Print an item.
	itcl::body PrinterDriver::TextOnly::print { ITEM_OBJECT_ARG } \
	{
		## Resize the item according to the paper width and any height.
		$interpreter eval $ITEM_OBJECT_ARG configure -width $paper_width -height $paper_height 

		## Print the primitives of the item.
		append raw_print_data [ convertPrimitives [ $interpreter eval $ITEM_OBJECT_ARG draw ] ]
	}
}

