#
# This script is part of TkKasse. You may use, modify and distribute it under the terms
# of the GNU General Public License, Version 2.
# See details at http://www.fsf.org/licenses/gpl.txt
# (C) by Jan Kandziora <tkkasse@users.sf.net> 
#


## Configuration and locale path setting:
set CONFIGURATION_PATH "/etc/opt/tkkasse"
set CR_AUTOSAVE_HARDCODED "/tmp/cr-autosave"
set LOCALE_PATH [ file join $env(LIBPATH) i18n ]


## Global procedures
## Make log entry.
proc log { LOGLEVEL_ARG MESSAGE_ARG } \
{
	global LOGLEVEL env

	## In debug loglevel, log all messages to stderr (debug file /var/log/tkkassed) additionally.
	if { $LOGLEVEL == "debug" } \
	{
		puts stderr "[ clock format [ clock seconds ] -format {%b %e %T} ] $env(HOSTNAME) tkkassed: $MESSAGE_ARG"
	}

	## Log all non-debug messages to syslog.
	if { $LOGLEVEL_ARG != "debug" } \
	{
		exec -- logger -p "daemon.$LOGLEVEL_ARG" -t tkkassed -- $MESSAGE_ARG
	}
}

## Handle background errors. 
proc bgerror { MESSAGE_ARG } \
{
	global errorCode errorInfo
	log crit "Unhandled background error: $errorCode\n$errorInfo"
}

## Backup all cash register data before cleaning up the cash register.
proc backupCashRegister {KEY} \
{
	global DATA_PATH errorInfo

	if [ catch \
		{
			set FD [ open "| gzip -c >$DATA_PATH/cr-$KEY.gz" "w" ]
			puts $FD [ getConfiguration TkKasse::Bill TkKasse::CashRegister ]
			close $FD
		} ] \
	{
		log crit "FATAL: Cash register data could *not* be backed up. Exiting!"
		log debug $errorInfo
		exit
	} \
	{
		log info "Cash register data backed up."
	}
	
}

## Clean up at exit
rename exit tcl_exit
proc exit {} \
{
	global CR_AUTOSAVE_HARDCODED DATA_PATH errorInfo

	if [ catch \
		{
			set FD [ open "$DATA_PATH/cr-autosave" "w" ]
			puts $FD [ getConfiguration TkKasse::Bill TkKasse::CashRegister ]
			close $FD
		} ] \
	{
		log crit "CRITICAL: Cash register data could *not* be saved at default location $DATA_PATH/cr-autosave."
		log debug $errorInfo
		log crit "Trying to save cash register data at $CR_AUTOSAVE_HARDCODED..."
		if [ catch \
			{
				set FD [ open $CR_AUTOSAVE_HARDCODED "w" ]
				puts $FD [ getConfiguration TkKasse::Bill TkKasse::CashRegister ]
				close $FD
			} ] \
		{
			log crit "Cash register data could *not* be saved at secondary location $CR_AUTOSAVE_HARDCODED. You lost data!"
		} \
		{
			log crit "Cash register data saved at secondary location $CR_AUTOSAVE_HARDCODED. Copy it by hand to $DATA_PATH/cr-autosave to restore it."
		}
	} \
	{
		log info "Cash register data saved."
	}

	mainServer stop
	log info "Stopped."
	tcl_exit
}



## Start tkkassed.
set LOGLEVEL info 
log info "Starting..."
set LOGLEVEL debug 

## Load packages.

## Tclx defines a command "log" (logarithm). We don't need this command,
## so we can simply delete it in favor of "our" "log" command.
rename log our_log
if [ catch { package require Tclx } ] \
{
	our_log err "FATAL: Packages could *not* be loaded. Exiting."
	global errorInfo
	log debug $errorInfo
	tcl_exit
}
catch { rename log {} }
rename our_log log


if [ catch \
	{
		package require Itcl
		package require msgcat
		package require crypt 1.0
	} ] \
{
	log err "FATAL: Packages could *not* be loaded. Exiting."
	global errorInfo
	log debug $errorInfo
	tcl_exit
}

## Load main program.
if [ catch \
	{
		source [ file join $env(LIBPATH) Utilities.tcl ]
		source [ file join $env(LIBPATH) Server.tcl ]
		source [ file join $env(LIBPATH) Connection.tcl ]
		source [ file join $env(LIBPATH) UserManager.tcl ]
		source [ file join $env(LIBPATH) TkKasseServer.tcl ]
		source [ file join $env(LIBPATH) TkKasseConnection.tcl ]
		source [ file join $env(LIBPATH) TkKasseUserManager.tcl ]
		source [ file join $env(LIBPATH) TkKasseArticleManager.tcl ]
		source [ file join $env(LIBPATH) TkKasseCashRegister.tcl ]
		source [ file join $env(LIBPATH) TkKassePrintManager.tcl ]
		source [ file join $env(LIBPATH) TkKassePrintFilter.tcl ]
		source [ file join $env(LIBPATH) TkKassePrinterDriver.tcl ]
		source [ file join $env(LIBPATH) TkKassePrintSpooler.tcl ]
	} ] \
{
	log err "FATAL: Main program could *not* be loaded. Exiting."
	log debug $errorInfo
	tcl_exit
}

## Trap normal program termination.
signal trap SIGTERM exit

## Ignore some other signals.
signal ignore SIGINT 
signal ignore SIGQUIT

## Read configuration files.
if [ catch \
	{
		source "$CONFIGURATION_PATH/settings"
		source "$CONFIGURATION_PATH/articles"
		source "$CONFIGURATION_PATH/printers"
		## Important note: The following one should be readable only for user/group tkkassed!
		source "$CONFIGURATION_PATH/users"
	} ] \
{
	log err "FATAL: Configuration could *not* be loaded. Exiting."
	log debug $errorInfo
	tcl_exit
} \
{
	log info "Configuration loaded."
}

## Set a special locale if specified in the configuration.
## Otherwise the system default locale is used.
if [ catch { set LOCALE } ] {} \
{
	msgcat::mclocale $LOCALE
	log info "Locale set to $LOCALE."
}

## Apply message catalog for internationalization.
if [ msgcat::mcload $LOCALE_PATH ] \
{
	log info "Localisation data loaded."
} \
{
	log err "FATAL: Localisation data could *not* be loaded. Exiting."
	log debug $errorInfo
	tcl_exit
}

## Read the automatically saved contents of last tkkassed session.
if [ catch \
	{
		## Important note: This file should be readable only for user/group tkkassed!
		source "$DATA_PATH/cr-autosave"
	} ] \
{
	log warn "WARNING: Cash register data could *not* be restored. Starting with empty cash register."
	log debug $errorInfo 
} \
{
	log info "Cash register data restored."
}

## If there was no cash register defined, define a new one.
if [ catch { ::TkKasse::CashRegister ::mainCashRegister -article_manager ::mainArticleManager -serial_number_counter "[ clock seconds ]/1" } ] {} \
{
	log info "Empty cash register created."
}

## Create a server object.
TkKasse::Server mainServer \
	-accept_port $SERVER_PORT \
	-user_manager mainUserManager \
	-cash_register mainCashRegister \
	-print_manager mainPrintManager \
	-connection_class TkKasse::Connection

## Allow clients to connect.
mainServer start

## Now wait for them to do.
vwait endless

